/*
libfawk - A function-only AWK dialect

Copyright (c) 2017..2020 Tibor 'Igor2' Palinkas. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  * 1. Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
  * 3. Neither the name of the copyright holder nor the names of its
    contributors may be used to endorse or promote products derived from this
    software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Project page: http://repo.hu/projects/libfawk
Source code: svn://repo.hu/libfawk/trunk
Contact the author: http://igor2.repo.hu/contact.html
*/

#include <stdlib.h>
#include <string.h>
#include "str.h"

/* warning: evaulates len twice */
#define FAWK_STR_ALLOC(len) \
	fawk_malloc(ctx, sizeof(fawk_str_t) + (len)); if (res == NULL) return NULL; \
	res->refco = 1; \
	res->used = res->alloced = (len);

FAWK_API fawk_str_t *fawk_str_new_from_literal(fawk_ctx_t *ctx, const char *s, size_t len_limit)
{
	size_t slen = strlen(s), len = (len_limit == -1) ? slen : ((slen < len_limit) ? slen : len_limit);
	fawk_str_t *res = FAWK_STR_ALLOC(len);
	memcpy(res->str, s, len);
	res->str[len] = '\0';
	return res;
}

FAWK_API fawk_str_t *fawk_str_clone(fawk_ctx_t *ctx, fawk_str_t *src, size_t enlarge)
{
	fawk_str_t *res = FAWK_STR_ALLOC(src->used + enlarge);
	memcpy(res->str, src->str, src->used+1);
	res->num = src->num;
	return res;
}

FAWK_API fawk_str_t *fawk_str_dup(fawk_ctx_t *ctx, fawk_str_t *src)
{
	src->refco++;
	if (src->refco == 0) {
		src->refco--;
		return fawk_str_clone(ctx, src, 0);
	}
	return src;
}

FAWK_API void fawk_str_free(fawk_ctx_t *ctx, fawk_str_t *src)
{
	fawk_assert(src->refco > 0);
	src->refco--;
	if (src->refco == 0) {
		FAWK_PKG_CALL(ctx, p->str_free_cb, (p, ctx, src));
		fawk_free(ctx, src);
	}
}

FAWK_API fawk_str_t *fawk_str_concat(fawk_ctx_t *ctx, fawk_str_t *s1, fawk_str_t *s2)
{
	fawk_str_t *res = FAWK_STR_ALLOC(s1->used + s2->used);
	memcpy(res->str, s1->str, s1->used);
	memcpy(res->str+s1->used, s2->str, s2->used+1);
	fawk_str_free(ctx, s1);
	fawk_str_free(ctx, s2);
	return res;
}
